from __future__ import annotations

from abc import ABC
from datetime import datetime
from typing import Union, List, TYPE_CHECKING, TypeVar

from pymeos_cffi import *
from spans import intrange, floatrange

from ..temporal import Temporal

if TYPE_CHECKING:
    from ..boxes import TBox
    from ..time import TimestampSet, Period, PeriodSet
    from .tfloat import TFloat

TBase = TypeVar('TBase', int, float)
TG = TypeVar('TG', 'TNumber[int]', 'TNumber[float]')
TI = TypeVar('TI', 'TInstant[int]', 'TInstant[float]')
TS = TypeVar('TS', 'TSequence[int]', 'TSequence[float]')
TSS = TypeVar('TSS', 'TSequenceSet[int]', 'TSequenceSet[float]')


class TNumber(Temporal[TBase, TG, TI, TS, TSS], ABC):

    def is_adjacent(self, other: Union[TBox, TNumber, floatrange, intrange,
                                       Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, TBox):
            return adjacent_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return adjacent_tnumber_tnumber(self._inner, other._inner)
        elif isinstance(other, floatrange):
            return adjacent_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, intrange):
            return adjacent_tnumber_span(self._inner, intrange_to_intspan(other))
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[TBox, TNumber, floatrange, intrange,
                                               Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(container, TBox):
            return contained_tnumber_tbox(self._inner, container._inner)
        elif isinstance(container, TNumber):
            return contained_tnumber_tnumber(self._inner, container._inner)
        elif isinstance(container, floatrange):
            return contained_tnumber_span(self._inner, floatrange_to_floatspan(container))
        elif isinstance(container, intrange):
            return contained_tnumber_span(self._inner, intrange_to_intspan(container))
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[TBox, TNumber, floatrange, intrange,
                                      Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(content, TBox):
            return contains_tnumber_tbox(self._inner, content._inner)
        elif isinstance(content, TNumber):
            return contains_tnumber_tnumber(self._inner, content._inner)
        elif isinstance(content, floatrange):
            return contains_tnumber_span(self._inner, floatrange_to_floatspan(content))
        elif isinstance(content, intrange):
            return contains_tnumber_span(self._inner, intrange_to_intspan(content))
        else:
            return super().contains(content)

    def is_left(self, other: Union[int, float, intrange, floatrange, TBox, TNumber]) -> bool:
        if isinstance(other, int):
            return left_tint_int(self._inner, other)
        elif isinstance(other, float):
            return left_tfloat_float(self._inner, other)
        elif isinstance(other, intrange):
            return left_tnumber_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            return left_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, TBox):
            return left_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return left_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other: Union[int, float, intrange, floatrange, TBox, TNumber]) -> bool:
        if isinstance(other, int):
            return overleft_tint_int(self._inner, other)
        elif isinstance(other, float):
            return overleft_tfloat_float(self._inner, other)
        elif isinstance(other, intrange):
            return overleft_tnumber_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            return overleft_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, TBox):
            return overleft_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overleft_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def overlaps(self, other: Union[TBox, TNumber, floatrange, intrange,
                                    Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, TBox):
            return overlaps_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overlaps_tnumber_tnumber(self._inner, other._inner)
        elif isinstance(other, floatrange):
            return overlaps_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, intrange):
            return overlaps_tnumber_span(self._inner, intrange_to_intspan(other))
        else:
            return super().overlaps(other)

    def is_over_or_right(self, other: Union[int, float, intrange, floatrange, TBox, TNumber]) -> bool:
        if isinstance(other, int):
            return overright_tint_int(self._inner, other)
        elif isinstance(other, float):
            return overright_tfloat_float(self._inner, other)
        elif isinstance(other, intrange):
            return overright_tnumber_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            return overright_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, TBox):
            return overright_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overright_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other: Union[int, float, intrange, floatrange, TBox, TNumber]) -> bool:
        if isinstance(other, int):
            return right_tint_int(self._inner, other)
        elif isinstance(other, float):
            return right_tfloat_float(self._inner, other)
        elif isinstance(other, intrange):
            return right_tnumber_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            return right_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, TBox):
            return right_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return right_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_same(self, other: Union[TBox, TNumber, floatrange, intrange,
                                   Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, TBox):
            return same_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return same_tnumber_tnumber(self._inner, other._inner)
        elif isinstance(other, floatrange):
            return same_tnumber_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, intrange):
            return same_tnumber_span(self._inner, intrange_to_intspan(other))
        else:
            return super().is_same(other)

    def is_before(self, other: Union[TBox, TNumber]) -> bool:
        if isinstance(other, TBox):
            return before_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return before_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_before(self, other: Union[TBox, TNumber]) -> bool:
        if isinstance(other, TBox):
            return overbefore_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overbefore_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_after(self, other: Union[TBox, TNumber]) -> bool:
        if isinstance(other, TBox):
            return after_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return after_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_after(self, other: Union[TBox, TNumber]) -> bool:
        if isinstance(other, TBox):
            return overafter_tnumber_tbox(self._inner, other._inner)
        elif isinstance(other, TNumber):
            return overafter_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def at(self, other: Union[intrange, floatrange, List[intrange], List[floatrange], TBox,
                              datetime, TimestampSet, Period, PeriodSet]) -> TG:
        from ..boxes import TBox
        if isinstance(other, intrange):
            result = tnumber_at_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            result = tnumber_at_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, list) and isinstance(other[0], intrange):
            spans = [intrange_to_intspan(ir) for ir in other]
            result = tnumber_at_spans(self._inner, spans, len(spans))
        elif isinstance(other, list) and isinstance(other[0], floatrange):
            spans = [floatrange_to_floatspan(fr) for fr in other]
            result = tnumber_at_spans(self._inner, spans, len(spans))
        elif isinstance(other, TBox):
            result = tnumber_at_tbox(self._inner, other._inner)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[intrange, floatrange, List[intrange], List[floatrange], TBox,
                                 datetime, TimestampSet, Period, PeriodSet]) -> TG:
        if isinstance(other, intrange):
            result = tnumber_minus_span(self._inner, intrange_to_intspan(other))
        elif isinstance(other, floatrange):
            result = tnumber_minus_span(self._inner, floatrange_to_floatspan(other))
        elif isinstance(other, list) and isinstance(other[0], intrange):
            spans = [intrange_to_intspan(ir) for ir in other]
            result = tnumber_minus_spans(self._inner, spans, len(spans))
        elif isinstance(other, list) and isinstance(other[0], floatrange):
            spans = [floatrange_to_floatspan(fr) for fr in other]
            result = tnumber_minus_spans(self._inner, spans, len(spans))
        elif isinstance(other, TBox):
            result = tnumber_minus_tbox(self._inner, other._inner)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    def distance(self, other: Union[int, float, TNumber]) -> TFloat:
        if isinstance(other, int):
            result = distance_tfloat_float(self._inner, float(other))
        elif isinstance(other, float):
            result = distance_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = distance_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def nearest_approach_distance(self, other: Union[int, float, TNumber, TBox]) -> float:
        if isinstance(other, int):
            return nad_tfloat_float(self._inner, float(other))
        elif isinstance(other, float):
            return nad_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            return nad_tfloat_tfloat(self._inner, other._inner)
        elif isinstance(other, TBox):
            return nad_tnumber_tbox(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def add(self, other: Union[int, float, TNumber]) -> TNumber:
        if isinstance(other, int):
            result = add_tint_int(self._inner, other)
        elif isinstance(other, float):
            result = add_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = add_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def radd(self, other: Union[int, float]) -> TNumber:
        if isinstance(other, int):
            result = add_int_tint(other, self._inner)
        elif isinstance(other, float):
            result = add_float_tfloat(other, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def sub(self, other: Union[int, float, TNumber]) -> TNumber:
        if isinstance(other, int):
            result = sub_tint_int(self._inner, other)
        elif isinstance(other, float):
            result = sub_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = sub_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rsub(self, other: Union[int, float]) -> TNumber:
        if isinstance(other, int):
            result = sub_int_tint(other, self._inner)
        elif isinstance(other, float):
            result = sub_float_tfloat(other, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def mul(self, other: Union[int, float, TNumber]) -> TNumber:
        if isinstance(other, int):
            result = mult_tint_int(self._inner, other)
        elif isinstance(other, float):
            result = mult_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = mult_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rmul(self, other: Union[int, float]) -> TNumber:
        if isinstance(other, int):
            result = mult_int_tint(other, self._inner)
        elif isinstance(other, float):
            result = mult_float_tfloat(other, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def div(self, other: Union[int, float, TNumber]) -> TNumber:
        if isinstance(other, int):
            result = div_tint_int(self._inner, other)
        elif isinstance(other, float):
            result = div_tfloat_float(self._inner, other)
        elif isinstance(other, TNumber):
            result = div_tnumber_tnumber(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def rdiv(self, other: Union[int, float]) -> TNumber:
        if isinstance(other, int):
            result = div_int_tint(other, self._inner)
        elif isinstance(other, float):
            result = div_float_tfloat(other, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def integral(self) -> float:
        return tnumber_integral(self._inner)

    def time_weighted_average(self) -> float:
        return tnumber_twavg(self._inner)

    def __add__(self, other):
        return self.add(other)

    def __radd__(self, other):
        return self.radd(other)

    def __sub__(self, other):
        return self.sub(other)

    def __rsub__(self, other):
        return self.rsub(other)

    def __mul__(self, other):
        return self.mul(other)

    def __rmul__(self, other):
        return self.rmul(other)

    def __truediv__(self, other):
        return self.div(other)

    def __rtruediv__(self, other):
        return self.rdiv(other)



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################
from __future__ import annotations

from typing import Optional, Union, List

from postgis import Geometry
from pymeos_cffi import *
from shapely.geometry.base import BaseGeometry

from ..main import TPoint
from ..time import *


class STBox:
    __slots__ = ['_inner']

    def __init__(self, string: Optional[str] = None, *,
                 xmin: Optional[Union[str, float]] = None, xmax: Optional[Union[str, float]] = None,
                 ymin: Optional[Union[str, float]] = None, ymax: Optional[Union[str, float]] = None,
                 zmin: Optional[Union[str, float]] = None, zmax: Optional[Union[str, float]] = None,
                 tmin: Optional[Union[str, datetime]] = None, tmax: Optional[Union[str, datetime]] = None,
                 tmin_inc: bool = True, tmax_inc: bool = True,
                 geodetic: bool = False, srid: Optional[int] = None,
                 _inner=None):

        assert (_inner is not None) or (string is not None) != (
                (xmin is not None and xmax is not None and ymin is not None and ymax is not None) or
                (tmin is not None and tmax is not None)), \
            "Either string must be not None or at least a bound pair (xmin/max and ymin/max, or tmin/max)" \
            " must be not None"

        if _inner is not None:
            self._inner = _inner
        elif string is not None:
            self._inner = stbox_in(string)
        else:
            period = None
            hast = tmin is not None and tmax is not None
            hasx = xmin is not None and xmax is not None and ymin is not None and ymax is not None
            hasz = zmin is not None and zmax is not None
            if hast:
                period = Period(lower=tmin, upper=tmax, lower_inc=tmin_inc, upper_inc=tmax_inc)._inner
            self._inner = stbox_make(period, hasx, hasz, geodetic, srid or 0, float(xmin or 0), float(xmax or 0),
                                     float(ymin or 0), float(ymax or 0), float(zmin or 0), float(zmax or 0))

    @staticmethod
    def from_hexwkb(hexwkb: str) -> STBox:
        result = stbox_from_hexwkb(hexwkb)
        return STBox(_inner=result)

    def as_hexwkb(self) -> str:
        return stbox_as_hexwkb(self._inner, -1)[0]

    @staticmethod
    def from_space(value: Geometry) -> STBox:
        return STBox.from_geometry(value)

    @staticmethod
    def from_geometry(geom: Geometry) -> STBox:
        gs = gserialized_in(geom.to_ewkb(), -1)
        return STBox(_inner=geo_to_stbox(gs))

    @staticmethod
    def from_time(time: Time) -> STBox:
        if isinstance(time, datetime):
            result = timestamp_to_stbox(datetime_to_timestamptz(time))
        elif isinstance(time, TimestampSet):
            result = timestampset_to_stbox(time)
        elif isinstance(time, Period):
            result = period_to_stbox(time)
        elif isinstance(time, PeriodSet):
            result = periodset_to_stbox(time)
        else:
            raise TypeError(f'Operation not supported with type {time.__class__}')
        return STBox(_inner=result)

    @staticmethod
    def from_expanding_bounding_box(value: Union[Geometry, TPoint], expansion: float):
        if isinstance(value, Geometry):
            gs = gserialized_in(value.to_ewkb(), -1)
            result = geo_expand_spatial(gs, expansion)
        elif isinstance(value, TPoint):
            result = tpoint_expand_spatial(value._inner, expansion)
        else:
            raise TypeError(f'Operation not supported with type {value.__class__}')
        return STBox(_inner=result)

    @staticmethod
    def from_space_time(value: Geometry, time: Union[datetime, Period]) -> STBox:
        return STBox.from_geometry_time(value, time)

    @staticmethod
    def from_geometry_time(geometry: Geometry, time: Union[datetime, Period]) -> STBox:
        gs = gserialized_in(geometry.to_ewkb(), -1)
        if isinstance(time, datetime):
            result = geo_timestamp_to_stbox(gs, datetime_to_timestamptz(time))
        elif isinstance(time, Period):
            result = geo_period_to_stbox(gs, time._inner)
        else:
            raise TypeError(f'Operation not supported with types {geometry.__class__} and {time.__class__}')
        return STBox(_inner=result)

    @staticmethod
    def from_tpoint(temporal: TPoint) -> STBox:
        return STBox(_inner=tpoint_to_stbox(temporal._inner))

    def tile(self, size: float, duration: Optional[Union[timedelta, str]] = None,
             origin: Optional[Union[BaseGeometry, Geometry]] = None,
             start: Union[datetime, str, None] = None) -> List[List[List[List[STBox]]]]:
        dt = timedelta_to_interval(duration) if isinstance(duration, timedelta) \
            else pg_interval_in(duration, -1) if isinstance(duration, str) \
            else None
        st = datetime_to_timestamptz(start) if isinstance(start, datetime) \
            else pg_timestamptz_in(start, -1) if isinstance(start, str) \
            else datetime_to_timestamptz(self.tmin) if self.has_t \
            else 0
        gs = geometry_to_gserialized(origin) if origin is not None \
            else gserialized_in('Point(0 0 0)', -1)
        tiles, dimensions = stbox_tile_list(self._inner, size, dt, gs, st)
        x_size = dimensions[0] or 1
        y_size = dimensions[1] or 1
        z_size = dimensions[2] or 1
        t_size = dimensions[3] or 1
        x_factor = y_size * z_size * t_size
        y_factor = z_size * t_size
        z_factor = t_size
        return [[[[STBox(_inner=tiles + x * x_factor + y * y_factor + z * z_factor + t) for t in range(t_size)]
                  for z in range(z_size)] for y in range(y_size)] for x in range(x_size)]

    def tile_flat(self, size: float, duration: Optional[Union[timedelta, str]] = None,
                  origin: Optional[Union[BaseGeometry, Geometry]] = None,
                  start: Union[datetime, str, None] = None) -> List[STBox]:
        boxes = self.tile(size, duration, origin, start)
        return [b
                for x in boxes
                for y in x
                for z in y
                for b in z
                ]

    def to_geometry(self, precision: int = 5) -> BaseGeometry:
        return gserialized_to_shapely_geometry(stbox_to_geo(self._inner), precision)

    def to_period(self) -> Period:
        return Period(_inner=stbox_to_period(self._inner))

    @property
    def has_xy(self):
        return stbox_hasx(self._inner)

    @property
    def has_z(self):
        return stbox_hasz(self._inner)

    @property
    def has_t(self):
        return stbox_hast(self._inner)

    @property
    def geodetic(self):
        """
        Is the box is geodetic?
        """
        return stbox_isgeodetic(self._inner)

    @property
    def xmin(self):
        """
        Minimum X
        """
        return stbox_xmin(self._inner)

    @property
    def ymin(self):
        """
        Minimum Y
        """
        return stbox_ymin(self._inner)

    @property
    def zmin(self):
        """
        Minimum Z
        """
        return stbox_zmin(self._inner)

    @property
    def tmin(self):
        """
        Minimum T
        """
        return timestamptz_to_datetime(stbox_tmin(self._inner))

    @property
    def xmax(self):
        """
        Maximum X
        """
        return stbox_xmax(self._inner)

    @property
    def ymax(self):
        """
        Maximum Y
        """
        return stbox_ymax(self._inner)

    @property
    def zmax(self):
        """
        Maximum Z
        """
        return stbox_zmax(self._inner)

    @property
    def tmax(self):
        """
        Maximum T
        """
        return timestamptz_to_datetime(stbox_tmax(self._inner))

    @property
    def srid(self):
        """
        SRID of the geographic coordinates
        """
        return self._inner.srid

    def set_srid(self, value: int) -> STBox:
        return STBox(_inner=stbox_set_srid(self._inner, value))

    def expand(self, other: Union[STBox, float, timedelta]) -> STBox:
        if isinstance(other, STBox):
            result = stbox_copy(self._inner)
            stbox_expand(other._inner, result)
        elif isinstance(other, float):
            result = stbox_expand_spatial(self._inner, other)
        elif isinstance(other, timedelta):
            result = stbox_expand_temporal(self._inner, timedelta_to_interval(other))
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return STBox(_inner=result)

    def shift_tscale(self, shift_delta: Optional[timedelta] = None, scale_delta: Optional[timedelta] = None):
        """
        Shift the spatio-temporal box by a time interval
        """
        assert shift_delta is not None or scale_delta is not None, 'shift and scale deltas must not be both None'
        stbox_shift_tscale(
            timedelta_to_interval(shift_delta) if shift_delta else None,
            timedelta_to_interval(scale_delta) if scale_delta else None,
            self._inner
        )

    def union(self, other: STBox, strict: bool = True) -> STBox:
        return STBox(_inner=union_stbox_stbox(self._inner, other._inner, strict))

    # TODO: Check returning None for empty intersection is the desired behaviour
    def intersection(self, other: STBox) -> Optional[STBox]:
        result = intersection_stbox_stbox(self._inner, other._inner)
        return STBox(_inner=result) if result else None

    def is_adjacent(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return adjacent_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return adjacent_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_contained_in(self, container: Union[STBox, TPoint]) -> bool:
        if isinstance(container, STBox):
            return contained_stbox_stbox(self._inner, container._inner)
        elif isinstance(container, TPoint):
            return contained_stbox_tpoint(self._inner, container._inner)
        else:
            raise TypeError(f'Operation not supported with type {container.__class__}')

    def contains(self, content: Union[STBox, TPoint]) -> bool:
        if isinstance(content, STBox):
            return contains_stbox_stbox(self._inner, content._inner)
        elif isinstance(content, TPoint):
            return contains_stbox_tpoint(self._inner, content._inner)
        else:
            raise TypeError(f'Operation not supported with type {content.__class__}')

    def overlaps(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overlaps_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overlaps_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_same(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return same_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return same_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_left(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return left_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return left_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overleft_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overleft_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return right_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return right_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overright_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overright_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_below(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return below_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return below_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_below(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overbelow_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overbelow_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_above(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return above_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return above_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_above(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overabove_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overabove_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_front(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return front_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return front_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_front(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overfront_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overfront_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_back(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return back_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return back_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_back(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overback_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overback_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_before(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return before_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return before_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_before(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overbefore_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overbefore_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_after(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return after_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return after_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_after(self, other: Union[STBox, TPoint]) -> bool:
        if isinstance(other, STBox):
            return overafter_stbox_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overafter_stbox_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def nearest_approach_distance(self, other: Union[Geometry, STBox]) -> float:
        if isinstance(other, Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return nad_stbox_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return nad_stbox_stbox(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def __add__(self, other):
        return self.union(other)

    def __mul__(self, other):
        return self.intersection(other)

    def __contains__(self, item):
        return self.contains(item)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return stbox_eq(self._inner, other._inner)
        return False

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return stbox_ne(self._inner, other._inner)
        return True

    def __cmp__(self, other):
        if isinstance(other, self.__class__):
            return stbox_cmp(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            return stbox_lt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __le__(self, other):
        if isinstance(other, self.__class__):
            return stbox_le(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __gt__(self, other):
        if isinstance(other, self.__class__):
            return stbox_gt(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __ge__(self, other):
        if isinstance(other, self.__class__):
            return stbox_ge(self._inner, other._inner)
        raise TypeError(f'Operation not supported with type {other.__class__}')

    def __copy__(self) -> STBox:
        inner_copy = stbox_copy(self._inner)
        return STBox(_inner=inner_copy)

    def __str__(self):
        return stbox_out(self._inner, 6)

    def __repr__(self):
        return (f'{self.__class__.__name__}'
                f'({self})')

    def plot_xy(self, *args, **kwargs):
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_xy(self, *args, **kwargs)

    def plot_xt(self, *args, **kwargs):
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_xt(self, *args, **kwargs)

    def plot_yt(self, *args, **kwargs):
        from ..plotters import BoxPlotter
        return BoxPlotter.plot_stbox_yt(self, *args, **kwargs)

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        return STBox(string=value)



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################
from __future__ import annotations

from abc import ABC
from typing import Optional, List, Union, TYPE_CHECKING, Set

from pymeos_cffi import *
from spans.types import floatrange, intrange

from .tnumber import TNumber
from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..time import *

if TYPE_CHECKING:
    from ..boxes import TBox
    from .tint import TInt


class TFloat(TNumber[float, 'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet'], ABC):
    BaseClass = float
    _parse_function = tfloat_in

    def is_adjacent(self, other: Union[int, float,
                                       TBox, TNumber, floatrange, intrange,
                                       Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int) or isinstance(other, float):
            return adjacent_tfloat_float(self._inner, float(other))
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[int, float,
                                               TBox, TNumber, floatrange, intrange,
                                               Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(container, int) or isinstance(container, float):
            return contained_tfloat_float(self._inner, float(container))
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[int, float,
                                      TBox, TNumber, floatrange, intrange,
                                      Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(content, int) or isinstance(content, float):
            return contains_tfloat_float(self._inner, float(content))
        else:
            return super().contains(content)

    def overlaps(self, other: Union[int, float,
                                    TBox, TNumber, floatrange, intrange,
                                    Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int) or isinstance(other, float):
            return overlaps_tfloat_float(self._inner, float(other))
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[int, float,
                                   TBox, TNumber, floatrange, intrange,
                                   Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int) or isinstance(other, float):
            return same_tfloat_float(self._inner, float(other))
        else:
            return super().is_same(other)

    def always_less(self, value: float) -> bool:
        return tfloat_always_lt(self._inner, value)

    def always_less_or_equal(self, value: float) -> bool:
        return tfloat_always_le(self._inner, value)

    def always_equal(self, value: float) -> bool:
        return tfloat_always_eq(self._inner, value)

    def always_not_equal(self, value: float) -> bool:
        return not tfloat_ever_eq(self._inner, value)

    def always_greater_or_equal(self, value: float) -> bool:
        return not tfloat_ever_lt(self._inner, value)

    def always_greater(self, value: float) -> bool:
        return not tfloat_ever_le(self._inner, value)

    def ever_less(self, value: float) -> bool:
        return tfloat_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: float) -> bool:
        return tfloat_ever_le(self._inner, value)

    def ever_equal(self, value: float) -> bool:
        return tfloat_ever_eq(self._inner, value)

    def ever_not_equal(self, value: float) -> bool:
        return not tfloat_always_eq(self._inner, value)

    def ever_greater_or_equal(self, value: float) -> bool:
        return not tfloat_always_lt(self._inner, value)

    def ever_greater(self, value: float) -> bool:
        return not tfloat_always_le(self._inner, value)

    def never_less(self, value: float) -> bool:
        return not tfloat_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: float) -> bool:
        return not tfloat_ever_le(self._inner, value)

    def never_equal(self, value: float) -> bool:
        return not tfloat_ever_eq(self._inner, value)

    def never_not_equal(self, value: float) -> bool:
        return tfloat_always_eq(self._inner, value)

    def never_greater_or_equal(self, value: float) -> bool:
        return tfloat_always_lt(self._inner, value)

    def never_greater(self, value: float) -> bool:
        return tfloat_always_le(self._inner, value)

    def temporal_less(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = tlt_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = tle_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = teq_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = tne_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = tge_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[int, float, Temporal]) -> Temporal:
        if isinstance(other, int) or isinstance(other, float):
            result = tgt_tfloat_float(self._inner, float(other))
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    def at(self, other: Union[int, float, List[float], List[int],
                              intrange, floatrange, List[intrange], List[floatrange], TBox,
                              datetime, TimestampSet, Period, PeriodSet]) -> TFloat:
        if isinstance(other, float) or isinstance(other, int):
            result = tfloat_at_value(self._inner, float(other))
        elif isinstance(other, list) and (isinstance(other[0], float) or isinstance(other[0], int)):
            result = tfloat_at_values(self._inner, [float(x) for x in other])
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[float, List[float],
                                 intrange, floatrange, List[intrange], List[floatrange], TBox,
                                 datetime, TimestampSet, Period, PeriodSet]) -> Temporal:
        if isinstance(other, float):
            result = tfloat_minus_value(self._inner, other)
        elif isinstance(other, list) and isinstance(other[0], float):
            result = tfloat_minus_values(self._inner, other)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    def to_tint(self) -> TInt:
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_to_tint(self._inner))

    def to_floatrange(self) -> floatrange:
        return floatspan_to_floatrange(tnumber_to_span(self._inner))

    @staticmethod
    def from_base(value: float, base: Temporal, interpolation: TInterpolation = TInterpolation.LINEAR) -> TFloat:
        result = tfloat_from_base(value, base._inner, interpolation)
        return Temporal._factory(result)

    @staticmethod
    def from_base_time(value: float, base: Time, interpolation: TInterpolation = None) -> TFloat:
        if isinstance(base, datetime):
            return TFloatInst(_inner=tfloatinst_make(value, datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TFloatSeq(_inner=tfloatdiscseq_from_base_time(value, base._inner))
        elif isinstance(base, Period):
            return TFloatSeq(_inner=tfloatseq_from_base_time(value, base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TFloatSeqSet(_inner=tfloatseqset_from_base_time(value, base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TFloatSeqSet(string=value)
            else:
                return TFloatSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TFloatInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TFloatSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TFloatSeqSet(string=value)
            else:
                return TFloatSeq(string=value)
        raise Exception("ERROR: Could not parse temporal float value")

    @property
    def value_range(self) -> floatrange:
        """
        Range of values taken by the temporal value as defined by its minimum and maximum value
        """
        return self.to_floatrange()

    @property
    def value_ranges(self) -> List[floatrange]:
        spans, count = tfloat_spans(self._inner)
        return [floatspan_to_floatrange(spans[i]) for i in range(count)]

    @property
    def start_value(self) -> float:
        """
        Start value.
        """
        return tfloat_start_value(self._inner)

    @property
    def end_value(self) -> float:
        """
        End value.
        """
        return tfloat_end_value(self._inner)

    @property
    def value_set(self) -> Set[float]:
        """
        List of distinct values.
        """
        values, count = tfloat_values(self._inner)
        return {values[i] for i in range(count)}

    @property
    def min_value(self) -> float:
        """
        Minimum value.
        """
        return tfloat_min_value(self._inner)

    @property
    def max_value(self) -> float:
        """
        Maximum value.
        """
        return tfloat_max_value(self._inner)

    def value_at_timestamp(self, timestamp) -> float:
        """
        Value at timestamp.
        """
        return tfloat_value_at_timestamp(self._inner, datetime_to_timestamptz(timestamp), True)

    def to_str(self, max_decimals=5) -> str:
        return tfloat_out(self._inner, max_decimals)

    def value_split(self, start: float, size: float) -> List[Temporal]:
        tiles, new_count = tfloat_value_split(self._inner, size, start)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(tiles[i]) for i in range(new_count)]

    def time_value_split(self, value_start: float, value_size: float, time_start: Union[str, datetime],
                         duration: Union[str, timedelta]) -> List[Temporal]:
        st = datetime_to_timestamptz(time_start) if isinstance(time_start, datetime) \
            else pg_timestamptz_in(time_start, -1)
        dt = timedelta_to_interval(duration) if isinstance(duration, timedelta) else pg_interval_in(duration, -1)
        tiles, new_count = tfloat_value_time_split(self._inner, value_size, value_start, dt, st)
        return [Temporal._factory(tiles[i]) for i in range(new_count)]

    def to_degrees(self) -> TNumber:
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_degrees(self._inner))

    def derivative(self) -> TNumber:
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tfloat_derivative(self._inner))

    def __str__(self):
        return tfloat_out(self._inner, 6)

    def as_wkt(self, precision: int = 6) -> str:
        return tfloat_out(self._inner, precision)


class TFloatInst(TInstant[float, 'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet'], TFloat):
    _make_function = tfloatinst_make
    _cast_function = int

    def __init__(self, string: Optional[str] = None, *, value: Optional[Union[str, float]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp, _inner=_inner)


class TFloatSeq(TSequence[float, 'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet'], TFloat):
    ComponentClass = TFloatInst

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, TFloatInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR, normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc,
                         expandable=expandable, interpolation=interpolation, normalize=normalize, _inner=_inner)


class TFloatSeqSet(TSequenceSet[float, 'TFloat', 'TFloatInst', 'TFloatSeq', 'TFloatSeqSet'], TFloat):
    ComponentClass = TFloatSeq

    def __init__(self, string: Optional[str] = None, *, sequence_list: Optional[List[Union[str, TFloatSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list, normalize=normalize, _inner=_inner)



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################
from __future__ import annotations

from abc import ABC
from typing import Optional, Union, List, Set

from pymeos_cffi import *

from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..time import *


class TText(Temporal[str, 'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet'], ABC):
    BaseClass = str

    _parse_function = ttext_in

    def __init__(self, _inner) -> None:
        super().__init__()

    def at(self, other: Union[str, List[str], datetime, TimestampSet, Period, PeriodSet]) -> TText:
        if isinstance(other, str):
            result = ttext_at_value(self._inner, other)
        elif isinstance(other, list):
            result = ttext_at_values(self._inner, other)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[str, List[str], datetime, TimestampSet, Period, PeriodSet]) -> TText:
        if isinstance(other, str):
            result = ttext_minus_value(self._inner, other)
        elif isinstance(other, list):
            result = ttext_minus_values(self._inner, other)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    @staticmethod
    def from_base(value: str, base: Temporal) -> TText:
        result = ttext_from_base(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    def from_base_time(value: str, base: Time) -> TText:
        if isinstance(base, datetime):
            return TTextInst(_inner=ttextinst_make(value, datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TTextSeq(_inner=ttextdiscseq_from_base_time(value, base._inner))
        elif isinstance(base, Period):
            return TTextSeq(_inner=ttextseq_from_base_time(value, base._inner))
        elif isinstance(base, PeriodSet):
            return TTextSeqSet(_inner=ttextseqset_from_base_time(value, base._inner))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    @property
    def value_set(self) -> Set[str]:
        values, count = ttext_values(self._inner)
        return {text2cstring(values[i]) for i in range(count)}

    @property
    def min_value(self):
        return ttext_min_value(self._inner)

    @property
    def max_value(self):
        return ttext_max_value(self._inner)

    @property
    def start_value(self):
        return ttext_start_value(self._inner)

    @property
    def end_value(self):
        return ttext_end_value(self._inner)

    def upper(self) -> TText:
        return self.__class__(_inner=ttext_upper(self._inner))

    def lower(self) -> TText:
        return self.__class__(_inner=ttext_lower(self._inner))

    def concatenate(self, other: Union[str, TText], other_before: bool = False):
        if isinstance(other, str):
            result = textcat_ttext_text(self._inner, other) if not other_before \
                else textcat_text_ttext(other, self._inner)
        elif isinstance(other, TText):
            result = textcat_ttext_ttext(self._inner, other._inner) if not other_before \
                else textcat_ttext_ttext(other._inner, self._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return self.__class__(_inner=result)

    def always_less(self, value: str) -> bool:
        return ttext_always_lt(self._inner, value)

    def always_less_or_equal(self, value: str) -> bool:
        return ttext_always_le(self._inner, value)

    def always_equal(self, value: str) -> bool:
        return ttext_always_eq(self._inner, value)

    def always_not_equal(self, value: str) -> bool:
        return not ttext_ever_eq(self._inner, value)

    def always_greater_or_equal(self, value: str) -> bool:
        return not ttext_ever_lt(self._inner, value)

    def always_greater(self, value: str) -> bool:
        return not ttext_ever_le(self._inner, value)

    def ever_less(self, value: str) -> bool:
        return ttext_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: str) -> bool:
        return ttext_ever_le(self._inner, value)

    def ever_equal(self, value: str) -> bool:
        return ttext_ever_eq(self._inner, value)

    def ever_not_equal(self, value: str) -> bool:
        return not ttext_always_eq(self._inner, value)

    def ever_greater_or_equal(self, value: str) -> bool:
        return not ttext_always_lt(self._inner, value)

    def ever_greater(self, value: str) -> bool:
        return not ttext_always_le(self._inner, value)

    def never_less(self, value: str) -> bool:
        return not ttext_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: str) -> bool:
        return not ttext_ever_le(self._inner, value)

    def never_equal(self, value: str) -> bool:
        return not ttext_ever_eq(self._inner, value)

    def never_not_equal(self, value: str) -> bool:
        return ttext_always_eq(self._inner, value)

    def never_greater_or_equal(self, value: str) -> bool:
        return ttext_always_lt(self._inner, value)

    def never_greater(self, value: str) -> bool:
        return ttext_always_le(self._inner, value)

    def temporal_less(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = tlt_ttext_text(self._inner, other)
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = tle_ttext_text(self._inner, other)
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_equal(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = teq_ttext_text(self._inner, other)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = tne_ttext_text(self._inner, other)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = tge_ttext_text(self._inner, other)
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[str, Temporal]) -> Temporal:
        if isinstance(other, str):
            result = tgt_ttext_text(self._inner, other)
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    def __add__(self, other):
        return self.concatenate(other)

    def __radd__(self, other):
        return self.concatenate(other, True)

    def value_at_timestamp(self, timestamp):
        """
        Value at timestamp.
        """
        return ttext_value_at_timestamp(self._inner, datetime_to_timestamptz(timestamp), True)

    def __str__(self):
        return ttext_out(self._inner)

    def as_wkt(self):
        return ttext_out(self._inner)

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        if value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TTextInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TTextSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TTextSeqSet(string=value)
            else:
                return TTextSeq(string=value)
        raise Exception("ERROR: Could not parse temporal text value")


class TTextInst(TInstant[str, 'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet'], TText):
    _make_function = ttextinst_make
    _cast_function = str

    def __init__(self, string: Optional[str] = None, *, value: Optional[str] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp, _inner=_inner)


class TTextSeq(TSequence[str, 'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet'], TText):
    ComponentClass = TTextInst

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, TTextInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.STEPWISE, normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc,
                         expandable=expandable, interpolation=interpolation, normalize=normalize, _inner=_inner)


class TTextSeqSet(TSequenceSet[str, 'TText', 'TTextInst', 'TTextSeq', 'TTextSeqSet'], TText):
    ComponentClass = TTextSeq

    def __init__(self, string: Optional[str] = None, *, sequence_list: Optional[List[Union[str, TTextSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list, normalize=normalize, _inner=_inner)



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################

from __future__ import annotations

from abc import ABC
from typing import Optional, List, TYPE_CHECKING, Set, Tuple, Union, TypeVar

import postgis as pg
import shapely.geometry as shp
import shapely.geometry.base as shpb
from geopandas import GeoDataFrame
from pymeos_cffi import *

from .tbool import TBool
from .tfloat import TFloatSeqSet, TFloat
from ..temporal import Temporal, TInstant, TSequence, TSequenceSet, TInterpolation
from ..time import *

if TYPE_CHECKING:
    from ..boxes import STBox

TG = TypeVar('TG', bound='TPoint')
TI = TypeVar('TI', bound='TPointInst')
TS = TypeVar('TS', bound='TPointSeq')
TSS = TypeVar('TSS', bound='TPointSeqSet')
Self = TypeVar('Self', bound='TPoint')


class TPoint(Temporal[shp.Point, TG, TI, TS, TSS], ABC):

    def __init__(self, _inner) -> None:
        super().__init__()

    @property
    def srid(self):
        """
        Returns the SRID.
        """
        return tpoint_srid(self._inner)

    def set_srid(self: Self, srid: int) -> Self:
        return self.__class__(_inner=tpoint_set_srid(self._inner, srid))

    def values(self, precision: int = 6) -> List[shp.Point]:
        return [i.value(precision=precision) for i in self.instants]

    def start_value(self, precision: int = 6) -> shp.Point:
        return gserialized_to_shapely_point(tpoint_start_value(self._inner), precision)

    def end_value(self, precision: int = 6) -> shp.Point:
        return gserialized_to_shapely_point(tpoint_end_value(self._inner), precision)

    def value_set(self, precision: int = 6) -> Set[shp.Point]:
        values, count = tpoint_values(self._inner)
        return {gserialized_to_shapely_point(values[i], precision) for i in range(count)}

    def value_at_timestamp(self, timestamp: datetime, precision: int = 6) -> shp.Point:
        """
        Value at timestamp.
        """
        return gserialized_to_shapely_point(
            tpoint_value_at_timestamp(self._inner, datetime_to_timestamptz(timestamp), True)[0], precision)

    def simplify(self: Self, tolerance: float, synchronized: bool = False) -> Self:
        return self.__class__(_inner=temporal_simplify(self._inner, tolerance, synchronized))

    def length(self) -> float:
        return tpoint_length(self._inner)

    def cumulative_length(self) -> TFloat:
        result = tpoint_cumulative_length(self._inner)
        return Temporal._factory(result)

    def speed(self) -> TFloat:
        result = tpoint_speed(self._inner)
        return Temporal._factory(result)

    def x(self) -> TFloat:
        result = tpoint_get_coord(self._inner, 0)
        return Temporal._factory(result)

    def y(self) -> TFloat:
        result = tpoint_get_coord(self._inner, 1)
        return Temporal._factory(result)

    def z(self) -> TFloat:
        result = tpoint_get_coord(self._inner, 2)
        return Temporal._factory(result)

    def stboxes(self) -> List[STBox]:
        from ..boxes import STBox
        result, count = tpoint_stboxes(self._inner)
        return [STBox(_inner=result + i) for i in range(count)]

    def is_simple(self) -> bool:
        return tpoint_is_simple(self._inner)

    def make_simple(self) -> List[TPoint]:
        result, count = tpoint_make_simple(self._inner)
        from ..factory import _TemporalFactory
        return [_TemporalFactory.create_temporal(result[i]) for i in range(count)]

    def is_adjacent(self, other: Union[pg.Geometry, STBox, TPoint,
                                       Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return adjacent_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return adjacent_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return adjacent_tpoint_tpoint(self._inner, other._inner)
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[pg.Geometry, STBox, TPoint,
                                               Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        from ..boxes import STBox
        if isinstance(container, pg.Geometry):
            gs = gserialized_in(container.to_ewkb(), -1)
            return contained_tpoint_geo(self._inner, gs)
        elif isinstance(container, STBox):
            return contained_tpoint_stbox(self._inner, container._inner)
        elif isinstance(container, TPoint):
            return contained_tpoint_tpoint(self._inner, container._inner)
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[pg.Geometry, STBox, TPoint,
                                      Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        from ..boxes import STBox
        if isinstance(content, pg.Geometry):
            gs = gserialized_in(content.to_ewkb(), -1)
            return contains_tpoint_geo(self._inner, gs)
        elif isinstance(content, STBox):
            return contains_tpoint_stbox(self._inner, content._inner)
        elif isinstance(content, TPoint):
            return contains_tpoint_tpoint(self._inner, content._inner)
        else:
            return super().contains(content)

    def overlaps(self, other: Union[pg.Geometry, STBox, TPoint,
                                    Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overlaps_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overlaps_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overlaps_tpoint_tpoint(self._inner, other._inner)
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[pg.Geometry, STBox, TPoint,
                                   Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return same_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return same_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return same_tpoint_tpoint(self._inner, other._inner)
        else:
            return super().is_same(other)

    def is_left(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return left_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return left_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return left_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_left(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overleft_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overleft_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overleft_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_right(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return right_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return right_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return right_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_right(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overright_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overright_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overright_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_below(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return below_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return below_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return below_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_below(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overbelow_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overbelow_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overbelow_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_above(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return above_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return above_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return above_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_above(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overabove_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overabove_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overabove_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_front(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return front_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return front_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return front_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_front(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overfront_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overfront_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overfront_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_back(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return back_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return back_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return back_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_back(self, other: Union[pg.Geometry, STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return overback_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return overback_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overback_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_before(self, other: Union[STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, STBox):
            return before_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return before_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_before(self, other: Union[STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, STBox):
            return overbefore_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overbefore_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_after(self, other: Union[STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, STBox):
            return after_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return after_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def is_over_or_after(self, other: Union[STBox, TPoint]) -> bool:
        from ..boxes import STBox
        if isinstance(other, STBox):
            return overafter_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return overafter_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def at(self, other: Union[pg.Geometry, List[pg.Geometry], shpb.BaseGeometry, List[shpb.BaseGeometry], STBox,
                              datetime, TimestampSet, Period, PeriodSet]) -> TG:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry) or isinstance(other, shpb.BaseGeometry):
            gs = geometry_to_gserialized(other)
            result = tpoint_at_geometry(self._inner, gs)
        elif isinstance(other, list):
            gss = [geometry_to_gserialized(gm) for gm in other]
            result = tpoint_at_values(self._inner, gss)
        elif isinstance(other, STBox):
            result = tpoint_at_stbox(self._inner, other._inner)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[pg.Geometry, List[pg.Geometry], shpb.BaseGeometry, List[shpb.BaseGeometry], STBox,
                                 datetime, TimestampSet, Period, PeriodSet]) -> TG:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry) or isinstance(other, shpb.BaseGeometry):
            gs = geometry_to_gserialized(other)
            result = tpoint_minus_geometry(self._inner, gs)
        elif isinstance(other, list):
            gss = [geometry_to_gserialized(gm) for gm in other]
            result = tpoint_minus_values(self._inner, gss)
        elif isinstance(other, STBox):
            result = tpoint_minus_stbox(self._inner, other._inner)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    def within_distance(self, other: Union[pg.Geometry, shpb.BaseGeometry, TPoint], distance: float) -> TBool:
        if isinstance(other, pg.Geometry) or isinstance(other, shpb.BaseGeometry):
            gs = geometry_to_gserialized(other)
            result = tdwithin_tpoint_geo(self._inner, gs, distance, False, False)
        elif isinstance(other, TPoint):
            result = tdwithin_tpoint_tpoint(self._inner, other._inner, distance, False, False)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def intersects(self, other: pg.Geometry) -> TBool:
        gs = gserialized_in(other.to_ewkb(), -1)
        result = tintersects_tpoint_geo(self._inner, gs, False, False)
        return Temporal._factory(result)

    def touches(self, other: pg.Geometry) -> TBool:
        gs = gserialized_in(other.to_ewkb(), -1)
        result = ttouches_tpoint_geo(self._inner, gs, False, False)
        return Temporal._factory(result)

    def is_contained(self, container: pg.Geometry) -> TBool:
        gs = gserialized_in(container.to_ewkb(), -1)
        result = tcontains_geo_tpoint(gs, self._inner, False, False)
        return Temporal._factory(result)

    def disjoint(self, other: pg.Geometry) -> TBool:
        gs = gserialized_in(other.to_ewkb(), -1)
        result = tdisjoint_tpoint_geo(self._inner, gs, False, False)
        return Temporal._factory(result)

    def is_ever_contained(self, container: pg.Geometry) -> bool:
        gs = gserialized_in(container.to_ewkb(), -1)
        return contains_geo_tpoint(gs, self._inner) == 1

    def is_ever_disjoint(self, other: Union[pg.Geometry, TPoint]) -> bool:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = disjoint_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = disjoint_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def is_ever_within_distance(self, other: Union[pg.Geometry, TPoint], distance: float) -> bool:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = dwithin_tpoint_geo(self._inner, gs, distance)
        elif isinstance(other, TPoint):
            result = dwithin_tpoint_tpoint(self._inner, other._inner, distance)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def ever_intersects(self, other: Union[pg.Geometry, TPoint]) -> bool:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = intersects_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = intersects_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return result == 1

    def ever_touches(self, other: pg.Geometry) -> bool:
        gs = gserialized_in(other.to_ewkb(), -1)
        return touches_tpoint_geo(gs, self._inner) == 1

    def distance(self, other: Union[pg.Geometry, TPoint]) -> TFloat:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = distance_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = distance_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def nearest_approach_distance(self, other: Union[pg.Geometry, STBox, TPoint]) -> float:
        from ..boxes import STBox
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            return nad_tpoint_geo(self._inner, gs)
        elif isinstance(other, STBox):
            return nad_tpoint_stbox(self._inner, other._inner)
        elif isinstance(other, TPoint):
            return nad_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')

    def nearest_approach_instant(self, other: Union[pg.Geometry, TPoint]) -> TI:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = nai_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = nai_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def shortest_line(self, other: Union[pg.Geometry, TPoint]) -> shpb.BaseGeometry:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = shortestline_tpoint_geo(self._inner, gs)
        elif isinstance(other, TPoint):
            result = shortestline_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return gserialized_to_shapely_geometry(result[0], 10)

    def bearing(self, other: Union[pg.Geometry, TPoint]) -> TFloat:
        if isinstance(other, pg.Geometry):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = bearing_tpoint_point(self._inner, gs, False)
        elif isinstance(other, TPoint):
            result = bearing_tpoint_tpoint(self._inner, other._inner)
        else:
            raise TypeError(f'Operation not supported with type {other.__class__}')
        return Temporal._factory(result)

    def azimuth(self) -> TFloatSeqSet:
        result = tpoint_azimuth(self._inner)
        return Temporal._factory(result)

    def time_weighted_centroid(self) -> shpb.BaseGeometry:
        return gserialized_to_shapely_geometry(tpoint_twcentroid(self._inner), 10)

    def tile(self, size: float, duration: Optional[Union[timedelta, str]] = None,
             origin: Optional[Union[shpb.BaseGeometry, pg.Geometry]] = None,
             start: Union[datetime, str, None] = None) -> List[List[List[List[TG]]]]:
        from ..boxes import STBox
        bbox = STBox.from_tpoint(self)
        tiles = bbox.tile(size, duration, origin, start)
        return [[[[self.at(tile) for tile in z_dim]
                  for z_dim in y_dim] for y_dim in x_dim] for x_dim in tiles]

    def tile_flat(self, size: float, duration: Optional[Union[timedelta, str]] = None,
                  origin: Optional[Union[shpb.BaseGeometry, pg.Geometry]] = None,
                  start: Union[datetime, str, None] = None) -> List[TG]:
        from ..boxes import STBox
        bbox = STBox.from_tpoint(self)
        tiles = bbox.tile_flat(size, duration, origin, start)
        return [x for x in (self.at(tile) for tile in tiles) if x]

    def as_geojson(self, option: int = 1, precision: int = 6, srs: Optional[str] = None) -> str:
        return gserialized_as_geojson(tpoint_trajectory(self._inner), option, precision, srs)

    def to_shapely_geometry(self, precision: int = 6) -> shpb.BaseGeometry:
        return gserialized_to_shapely_geometry(tpoint_trajectory(self._inner), precision)

    def to_dataframe(self) -> GeoDataFrame:
        data = {
            'time': self.timestamps,
            'geometry': [i.value() for i in self.instants]
        }
        return GeoDataFrame(data, crs=self.srid).set_index(keys=['time'])

    def __str__(self):
        return tpoint_out(self._inner, 6)

    def as_wkt(self, precision: int = 6):
        return tpoint_out(self._inner, precision)

    def as_ewkt(self, precision: int = 6):
        return tpoint_as_ewkt(self._inner, precision)


class TPointInst(TInstant[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):
    def value(self, precision: int = 6) -> shp.Point:
        return self.start_value(precision=precision)


class TPointSeq(TSequence[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):

    @staticmethod
    def from_arrays(t: List[Union[datetime, str]], x: List[float], y: List[float], z: Optional[List[float]] = None,
                    srid: int = 0, geodetic: bool = False, lower_inc: bool = True, upper_inc: bool = False,
                    interpolation: TInterpolation = TInterpolation.LINEAR, normalize: bool = True) -> TPointSeq:
        from ..factory import _TemporalFactory
        assert len(t) == len(x) == len(y)
        times = [datetime_to_timestamptz(ti) if isinstance(ti, datetime) else pg_timestamptz_in(ti, -1) for ti in t]
        return _TemporalFactory.create_temporal(
            tpointseq_make_coords(x, y, z, times, len(t), srid, geodetic, lower_inc, upper_inc, interpolation,
                                  normalize)
        )

    def plot(self, *args, **kwargs):
        from ..plotters import TemporalPointSequencePlotter
        return TemporalPointSequencePlotter.plot_xy(self, *args, **kwargs)


class TPointSeqSet(TSequenceSet[shpb.BaseGeometry, TG, TI, TS, TSS], TPoint[TG, TI, TS, TSS], ABC):

    @property
    def speed(self):
        return TFloatSeqSet(_inner=tpoint_speed(self._inner))

    def to_dataframe(self, precision: int = 6) -> GeoDataFrame:
        data = {
            'sequence': [i + 1 for i, seq in enumerate(self.sequences) for _ in range(seq.instants)],
            'time': [t for seq in self.sequences for t in seq.timestamps],
            'geometry': [v for seq in self.sequences for v in seq.values(precision=precision)]
        }
        return GeoDataFrame(data, crs=self.srid).set_index(keys=['sequence', 'time'])

    def plot(self, *args, **kwargs):
        from ..plotters import TemporalPointSequenceSetPlotter
        return TemporalPointSequenceSetPlotter.plot_xy(self, *args, **kwargs)


class TGeomPoint(TPoint['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet'], ABC):
    BaseClass = pg.Point
    _parse_function = tgeompoint_in

    @staticmethod
    def from_base(value: pg.Geometry, base: Temporal,
                  interpolation: TInterpolation = TInterpolation.LINEAR) -> TGeomPoint:
        gs = gserialized_in(value.to_ewkb(), -1)
        result = tgeompoint_from_base(gs, base._inner, interpolation)
        return Temporal._factory(result)

    @staticmethod
    def from_base_time(value: pg.Geometry, base: Time, interpolation: TInterpolation = None) -> TGeomPoint:
        gs = gserialized_in(value.to_ewkb(), -1)
        if isinstance(base, datetime):
            return TGeomPointInst(_inner=tgeompointinst_make(gs, datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TGeomPointSeq(_inner=tgeompointdiscseq_from_base_time(gs, base._inner))
        elif isinstance(base, Period):
            return TGeomPointSeq(_inner=tgeompointseq_from_base_time(gs, base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TGeomPointSeqSet(_inner=tgeompointseqset_from_base_time(gs, base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    def to_geographic(self) -> TGeogPoint:
        result = tgeompoint_tgeogpoint(self._inner, True)
        return Temporal._factory(result)

    def always_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeompoint_always_eq(self._inner, gs)

    def always_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeompoint_ever_eq(self._inner, gs)

    def ever_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeompoint_ever_eq(self._inner, gs)

    def ever_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeompoint_always_eq(self._inner, gs)

    def never_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeompoint_ever_eq(self._inner, gs)

    def never_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeompoint_always_eq(self._inner, gs)

    def temporal_equal(self, other: Union[pg.Point, Temporal]) -> Temporal:
        if isinstance(other, pg.Point):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = teq_tgeompoint_point(self._inner, gs)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[pg.Point, Temporal]) -> Temporal:
        if isinstance(other, pg.Point):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = tne_tgeompoint_point(self._inner, gs)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TGeomPointSeqSet(string=value)
            else:
                return TGeomPointSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TGeomPointInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TGeomPointSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TGeomPointSeqSet(string=value)
            else:
                return TGeomPointSeq(string=value)
        raise Exception("ERROR: Could not parse temporal point value")

    @property
    def hasz(self):
        """
        Does the temporal point has Z dimension?
        """
        return self.start_value().has_z


class TGeogPoint(TPoint['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet'], ABC):
    BaseClass = pg.Point
    _parse_function = tgeogpoint_in

    @staticmethod
    def from_base(value: pg.Geometry, base: Temporal,
                  interpolation: TInterpolation = TInterpolation.LINEAR) -> TGeogPoint:
        gs = gserialized_in(value.to_ewkb(), -1)
        result = tgeogpoint_from_base(gs, base._inner, interpolation)
        return Temporal._factory(result)

    @staticmethod
    def from_base_time(value: pg.Geometry, base: Time, interpolation: TInterpolation = None) -> TGeogPoint:
        gs = gserialized_in(value.to_ewkb(), -1)
        if isinstance(base, datetime):
            return TGeogPointInst(_inner=tgeogpointinst_make(gs, datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TGeogPointSeq(_inner=tgeogpointdiscseq_from_base_time(gs, base._inner))
        elif isinstance(base, Period):
            return TGeogPointSeq(_inner=tgeogpointseq_from_base_time(gs, base._inner, interpolation))
        elif isinstance(base, PeriodSet):
            return TGeogPointSeqSet(_inner=tgeogpointseqset_from_base_time(gs, base._inner, interpolation))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    def to_geometric(self) -> TGeogPoint:
        result = tgeompoint_tgeogpoint(self._inner, False)
        return Temporal._factory(result)

    def always_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeogpoint_always_eq(self._inner, gs)

    def always_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeogpoint_ever_eq(self._inner, gs)

    def ever_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeogpoint_ever_eq(self._inner, gs)

    def ever_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeogpoint_always_eq(self._inner, gs)

    def never_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return not tgeogpoint_ever_eq(self._inner, gs)

    def never_not_equal(self, value: pg.Geometry) -> bool:
        gs = gserialized_in(value.to_ewkb(), -1)
        return tgeogpoint_always_eq(self._inner, gs)

    def temporal_equal(self, other: Union[pg.Point, Temporal]) -> Temporal:
        if isinstance(other, pg.Point):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = teq_tgeogpoint_point(self._inner, gs)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[pg.Point, Temporal]) -> Temporal:
        if isinstance(other, pg.Point):
            gs = gserialized_in(other.to_ewkb(), -1)
            result = tne_tgeogpoint_point(self._inner, gs)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        if value.startswith('Interp=Stepwise;'):
            value1 = value.replace('Interp=Stepwise;', '')
            if value1[0] == '{':
                return TGeogPointSeqSet(string=value)
            else:
                return TGeogPointSeq(string=value)
        elif value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TGeogPointInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TGeogPointSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TGeogPointSeqSet(string=value)
            else:
                return TGeogPointSeq(string=value)
        raise Exception("ERROR: Could not parse temporal point value")

    @property
    def hasz(self):
        """
        Does the temporal point has Z dimension?
        """
        return self.start_value().has_z


class TGeomPointInst(TPointInst['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet'], TGeomPoint):
    _make_function = lambda *args: None
    _cast_function = lambda x: None

    def __init__(self, string: Optional[str] = None, *, point: Optional[Union[str, pg.Point]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, srid: Optional[int] = 0, _inner=None) -> None:
        super().__init__(string=string, value=point, timestamp=timestamp, _inner=_inner)
        if self._inner is None:
            self._inner = tgeompoint_in(f"SRID={srid};{point}@{timestamp}")


class TGeogPointInst(TPointInst['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet'], TGeogPoint):
    _make_function = lambda *args: None
    _cast_function = lambda x: None

    def __init__(self, string: Optional[str] = None, *,
                 point: Optional[Union[str, pg.Point, Tuple[float, float]]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, srid: Optional[int] = 0, _inner=None) -> None:
        super().__init__(string=string, value=point, timestamp=timestamp, _inner=_inner)
        if self._inner is None:
            p = f'POINT({point[0]} {point[1]})' if isinstance(point, tuple) else f'{point}'
            self._inner = tgeogpoint_in(f"SRID={srid};{p}@{timestamp}")


class TGeomPointSeq(TPointSeq['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet'], TGeomPoint):
    ComponentClass = TGeomPointInst

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, TGeomPointInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc,
                         expandable=expandable, interpolation=interpolation, normalize=normalize, _inner=_inner)


class TGeogPointSeq(TPointSeq['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet'], TGeogPoint):
    ComponentClass = TGeogPointInst

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, TGeogPointInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR, normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc,
                         expandable=expandable, interpolation=interpolation, normalize=normalize, _inner=_inner)


class TGeomPointSeqSet(TPointSeqSet['TGeomPoint', 'TGeomPointInst', 'TGeomPointSeq', 'TGeomPointSeqSet'], TGeomPoint):
    ComponentClass = TGeomPointSeq

    def __init__(self, string: Optional[str] = None, *, sequence_list: Optional[List[Union[str, TGeomPointSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list, normalize=normalize, _inner=_inner)


class TGeogPointSeqSet(TPointSeqSet['TGeogPoint', 'TGeogPointInst', 'TGeogPointSeq', 'TGeogPointSeqSet'], TGeogPoint):
    ComponentClass = TGeogPointSeq

    def __init__(self, string: Optional[str] = None, *, sequence_list: Optional[List[Union[str, TGeogPointSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list, normalize=normalize, _inner=_inner)



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..boxes import STBox
from ..main import TPoint


class TemporalPointExtentAggregator(BaseAggregator[TPoint, STBox]):
    _add_function = tpoint_extent_transfn

    @classmethod
    def _finish(cls, state) -> STBox:
        return STBox(_inner=state)



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..main import TText


class TemporalTextMaxAggregator(BaseAggregator[TText, TText]):
    _add_function = ttext_tmax_transfn


class TemporalTextMinAggregator(BaseAggregator[TText, TText]):
    _add_function = ttext_tmin_transfn



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################
from __future__ import annotations

from abc import ABC
from typing import Optional, Union, List, TYPE_CHECKING, Set

from pymeos_cffi import *
from spans.types import intrange, floatrange

from .tnumber import TNumber
from ..temporal import TInterpolation, Temporal, TInstant, TSequence, TSequenceSet
from ..time import *

if TYPE_CHECKING:
    from ..boxes import TBox
    from .tfloat import TFloat


class TInt(TNumber[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], ABC):
    BaseClass = int
    _parse_function = tint_in

    def is_adjacent(self, other: Union[int,
                                       TBox, TNumber, floatrange, intrange,
                                       Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int):
            return adjacent_tint_int(self._inner, other)
        else:
            return super().is_adjacent(other)

    def is_contained_in(self, container: Union[int,
                                               TBox, TNumber, floatrange, intrange,
                                               Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(container, int):
            return contained_tint_int(self._inner, container)
        else:
            return super().is_contained_in(container)

    def contains(self, content: Union[int,
                                      TBox, TNumber, floatrange, intrange,
                                      Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(content, int):
            return contains_tint_int(self._inner, content)
        else:
            return super().contains(content)

    def overlaps(self, other: Union[int,
                                    TBox, TNumber, floatrange, intrange,
                                    Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int):
            return overlaps_tint_int(self._inner, other)
        else:
            return super().overlaps(other)

    def is_same(self, other: Union[int,
                                   TBox, TNumber, floatrange, intrange,
                                   Period, PeriodSet, datetime, TimestampSet, Temporal]) -> bool:
        if isinstance(other, int):
            return same_tint_int(self._inner, other)
        else:
            return super().is_same(other)

    def always_less(self, value: int) -> bool:
        return tint_always_lt(self._inner, value)

    def always_less_or_equal(self, value: int) -> bool:
        return tint_always_le(self._inner, value)

    def always_equal(self, value: int) -> bool:
        return tint_always_eq(self._inner, value)

    def always_not_equal(self, value: int) -> bool:
        return not tint_ever_eq(self._inner, value)

    def always_greater_or_equal(self, value: int) -> bool:
        return not tint_ever_lt(self._inner, value)

    def always_greater(self, value: int) -> bool:
        return not tint_ever_le(self._inner, value)

    def ever_less(self, value: int) -> bool:
        return tint_ever_lt(self._inner, value)

    def ever_less_or_equal(self, value: int) -> bool:
        return tint_ever_le(self._inner, value)

    def ever_equal(self, value: int) -> bool:
        return tint_ever_eq(self._inner, value)

    def ever_not_equal(self, value: int) -> bool:
        return not tint_always_eq(self._inner, value)

    def ever_greater_or_equal(self, value: int) -> bool:
        return not tint_always_lt(self._inner, value)

    def ever_greater(self, value: int) -> bool:
        return not tint_always_le(self._inner, value)

    def never_less(self, value: int) -> bool:
        return not tint_ever_lt(self._inner, value)

    def never_less_or_equal(self, value: int) -> bool:
        return not tint_ever_le(self._inner, value)

    def never_equal(self, value: int) -> bool:
        return not tint_ever_eq(self._inner, value)

    def never_not_equal(self, value: int) -> bool:
        return tint_always_eq(self._inner, value)

    def never_greater_or_equal(self, value: int) -> bool:
        return tint_always_lt(self._inner, value)

    def never_greater(self, value: int) -> bool:
        return tint_always_le(self._inner, value)

    def temporal_less(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = tlt_tint_int(self._inner, other)
        else:
            return super().temporal_less(other)
        return Temporal._factory(result)

    def temporal_less_or_equal(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = tle_tint_int(self._inner, other)
        else:
            return super().temporal_less_or_equal(other)
        return Temporal._factory(result)

    def temporal_equal(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = teq_tint_int(self._inner, other)
        else:
            return super().temporal_equal(other)
        return Temporal._factory(result)

    def temporal_not_equal(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = tne_tint_int(self._inner, other)
        else:
            return super().temporal_not_equal(other)
        return Temporal._factory(result)

    def temporal_greater_or_equal(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = tge_tint_int(self._inner, other)
        else:
            return super().temporal_greater_or_equal(other)
        return Temporal._factory(result)

    def temporal_greater(self, other: Union[int, Temporal]) -> Temporal:
        if isinstance(other, int):
            result = tgt_tint_int(self._inner, other)
        else:
            return super().temporal_greater(other)
        return Temporal._factory(result)

    def at(self, other: Union[int, List[int],
                              intrange, floatrange, List[intrange], List[floatrange], TBox,
                              datetime, TimestampSet, Period, PeriodSet]) -> Temporal:
        if isinstance(other, int):
            result = tint_at_value(self._inner, other)
        elif isinstance(other, list) and isinstance(other[0], int):
            result = tint_at_values(self._inner, other)
        else:
            return super().at(other)
        return Temporal._factory(result)

    def minus(self, other: Union[int, List[int],
                                 intrange, floatrange, List[intrange], List[floatrange], TBox,
                                 datetime, TimestampSet, Period, PeriodSet]) -> Temporal:
        if isinstance(other, int):
            result = tint_minus_value(self._inner, other)
        elif isinstance(other, list) and isinstance(other[0], int):
            result = tint_minus_values(self._inner, other)
        else:
            return super().minus(other)
        return Temporal._factory(result)

    def nearest_approach_distance(self, other: Union[int, float, TNumber, TBox]) -> float:
        if isinstance(other, int):
            return nad_tint_int(self._inner, other)
        elif isinstance(other, TInt):
            return nad_tint_tint(self._inner, other._inner)
        else:
            return super(TInt, self).nearest_approach_distance(other)

    def to_tfloat(self) -> TFloat:
        from ..factory import _TemporalFactory
        return _TemporalFactory.create_temporal(tint_to_tfloat(self._inner))

    def to_intrange(self) -> intrange:
        return intspan_to_intrange(tnumber_to_span(self._inner))

    @staticmethod
    def from_base(value: int, base: Temporal) -> TInt:
        result = tint_from_base(value, base._inner)
        return Temporal._factory(result)

    @staticmethod
    def from_base_time(value: int, base: Time) -> TInt:
        if isinstance(base, datetime):
            return TIntInst(_inner=tintinst_make(value, datetime_to_timestamptz(base)))
        elif isinstance(base, TimestampSet):
            return TIntSeq(_inner=tintdiscseq_from_base_time(value, base._inner))
        elif isinstance(base, Period):
            return TIntSeq(_inner=tintseq_from_base_time(value, base._inner))
        elif isinstance(base, PeriodSet):
            return TIntSeqSet(_inner=tintseqset_from_base_time(value, base._inner))
        raise TypeError(f'Operation not supported with type {base.__class__}')

    @staticmethod
    def read_from_cursor(value, _=None):
        if not value:
            return None
        if value[0] != '{' and value[0] != '[' and value[0] != '(':
            return TIntInst(string=value)
        elif value[0] == '[' or value[0] == '(':
            return TIntSeq(string=value)
        elif value[0] == '{':
            if value[1] == '[' or value[1] == '(':
                return TIntSeqSet(string=value)
            else:
                return TIntSeq(string=value)
        raise Exception("ERROR: Could not parse temporal integer value")

    @property
    def value_range(self) -> intrange:
        """
        Range of values taken by the temporal value as defined by its minimum and maximum value
        """
        return self.to_intrange()

    @property
    def start_value(self) -> int:
        """
        Start value.
        """
        return tint_start_value(self._inner)

    @property
    def end_value(self) -> int:
        """
        End value.
        """
        return tint_end_value(self._inner)

    @property
    def value_set(self) -> Set[int]:
        """
        List of distinct values.
        """
        values, count = tint_values(self._inner)
        return {values[i] for i in range(count)}

    @property
    def min_value(self) -> int:
        """
        Minimum value.
        """
        return tint_min_value(self._inner)

    @property
    def max_value(self) -> int:
        """
        Maximum value.
        """
        return tint_max_value(self._inner)

    def value_at_timestamp(self, timestamp) -> int:
        """
        Value at timestamp.
        """
        return tint_value_at_timestamp(self._inner, datetime_to_timestamptz(timestamp), True)

    def value_split(self, start: int, size: int) -> List[Temporal]:
        tiles, new_count = tint_value_split(self._inner, size, start)
        return [Temporal._factory(tiles[i]) for i in range(new_count)]

    def time_value_split(self, value_start: int, value_size: int, time_start: Union[str, datetime],
                         duration: Union[str, timedelta]) -> List[Temporal]:
        st = datetime_to_timestamptz(time_start) if isinstance(time_start, datetime) \
            else pg_timestamptz_in(time_start, -1)
        dt = timedelta_to_interval(duration) if isinstance(duration, timedelta) else pg_interval_in(duration, -1)
        tiles, new_count = tint_value_time_split(self._inner, value_size, value_start, dt, st)
        return [Temporal._factory(tiles[i]) for i in range(new_count)]

    def __str__(self):
        return tint_out(self._inner)

    def as_wkt(self):
        return tint_out(self._inner)


class TIntInst(TInstant[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], TInt):
    _make_function = tintinst_make
    _cast_function = int

    def __init__(self, string: Optional[str] = None, *, value: Optional[Union[str, int]] = None,
                 timestamp: Optional[Union[str, datetime]] = None, _inner=None):
        super().__init__(string=string, value=value, timestamp=timestamp, _inner=_inner)


class TIntSeq(TSequence[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], TInt):
    ComponentClass = TIntInst

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, TIntInst]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.STEPWISE, normalize: bool = True, _inner=None):
        super().__init__(string=string, instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc,
                         expandable=expandable, interpolation=interpolation, normalize=normalize, _inner=_inner)


class TIntSeqSet(TSequenceSet[int, 'TInt', 'TIntInst', 'TIntSeq', 'TIntSeqSet'], TInt):
    ComponentClass = TIntSeq

    def __init__(self, string: Optional[str] = None, *, sequence_list: Optional[List[Union[str, TIntSeq]]] = None,
                 normalize: bool = True, _inner=None):
        super().__init__(string=string, sequence_list=sequence_list, normalize=normalize, _inner=_inner)



###############################################################################
#
# This MobilityDB code is provided under The PostgreSQL License.
#
# Copyright (c) 2019-2022, Université libre de Bruxelles and MobilityDB
# contributors
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose, without fee, and without a written 
# agreement is hereby granted, provided that the above copyright notice and
# this paragraph and the following two paragraphs appear in all copies.
#
# IN NO EVENT SHALL UNIVERSITE LIBRE DE BRUXELLES BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
# LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION,
# EVEN IF UNIVERSITE LIBRE DE BRUXELLES HAS BEEN ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.
#
# UNIVERSITE LIBRE DE BRUXELLES SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
# AN "AS IS" BASIS, AND UNIVERSITE LIBRE DE BRUXELLES HAS NO OBLIGATIONS TO 
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. 
#
###############################################################################
from __future__ import annotations

from abc import ABC
from typing import Optional, Union, List, Any, TypeVar, Type

from pymeos_cffi import *

from .interpolation import TInterpolation
from .temporal import Temporal

TBase = TypeVar('TBase')
TG = TypeVar('TG', bound='Temporal[Any]')
TI = TypeVar('TI', bound='TInstant[Any]')
TS = TypeVar('TS', bound='TSequence[Any]')
TSS = TypeVar('TSS', bound='TSequenceSet[Any]')
Self = TypeVar('Self', bound='TSequence[Any]')


class TSequence(Temporal[TBase, TG, TI, TS, TSS], ABC):
    """
    Abstract class for representing temporal values of sequence subtype.
    """

    def __init__(self, string: Optional[str] = None, *, instant_list: Optional[List[Union[str, Any]]] = None,
                 lower_inc: bool = True, upper_inc: bool = False, expandable: Union[bool, int] = False,
                 interpolation: TInterpolation = TInterpolation.LINEAR, normalize: bool = True, _inner=None):
        assert (_inner is not None) or ((string is not None) != (instant_list is not None)), \
            "Either string must be not None or instant_list must be not"
        if _inner is not None:
            self._inner = as_tsequence(_inner)
        elif string is not None:
            self._inner = as_tsequence(self.__class__._parse_function(string))
        else:
            self._instants = [x._inner if isinstance(x, self.ComponentClass) else self.__class__._parse_function(x) for
                              x in instant_list]
            count = len(self._instants)
            if not expandable:
                self._inner = tsequence_make(self._instants, count, lower_inc, upper_inc,
                                             interpolation.value, normalize)
            else:
                max_size = max(expandable, count) if isinstance(expandable, int) else 2 * count
                self._inner = tsequence_make_exp(self._instants, count, max_size, lower_inc, upper_inc,
                                                 interpolation.value, normalize)
        self._expandable_sequence = bool(expandable) or self._inner.maxcount > self._inner.count

    @property
    def _expandable(self) -> bool:
        return self._expandable_sequence

    @classmethod
    def from_instants(cls: Type[Self], instant_list: Optional[List[Union[str, Any]]],
                      lower_inc: bool = True, upper_inc: bool = False,
                      interpolation: TInterpolation = TInterpolation.LINEAR, normalize: bool = True) -> Self:
        return cls(instant_list=instant_list, lower_inc=lower_inc, upper_inc=upper_inc, interpolation=interpolation,
                   normalize=normalize)

    @property
    def lower_inc(self) -> bool:
        """
        Is the lower bound inclusive?
        """
        return self._inner.period.lower_inc

    @property
    def upper_inc(self) -> bool:
        """
        Is the upper bound inclusive?
        """
        return self._inner.period.upper_inc

    @property
    def num_sequences(self) -> int:
        """
        Number of sequences.
        """
        return 1

    @property
    def start_sequence(self: Self) -> Self:
        """
        Start sequence.
        """
        return self

    @property
    def end_sequence(self: Self) -> Self:
        """
        End sequence.
        """
        return self

    def sequence_n(self: Self, n) -> Self:
        """
        N-th sequence.
        """
        # 1-based
        if n == 1:
            return self
        else:
            raise Exception("ERROR: Out of range")

    @property
    def sequences(self: Self) -> List[Self]:
        """
        List of sequences.
        """
        return [self]

    def compact(self: Self) -> Self:
        result = tsequence_compact(self._inner)
        return Temporal._factory(result)

    def plot(self, *args, **kwargs):
        from ..plotters import TemporalSequencePlotter
        return TemporalSequencePlotter.plot(self, *args, **kwargs)



from typing import Union, List

from .point_sequence_plotter import TemporalPointSequencePlotter
from .. import TPointSeq
from ..main import TPointSeqSet


class TemporalPointSequenceSetPlotter:

    @staticmethod
    def plot_xy(sequence_set: Union[TPointSeqSet, List[TPointSeq]], *args, **kwargs):
        seqs = sequence_set.sequences if isinstance(sequence_set, TPointSeqSet) else sequence_set
        plot = TemporalPointSequencePlotter.plot_xy(seqs[0], *args, **kwargs)
        if 'color' not in kwargs:
            kwargs['color'] = plot[0].get_color()
        kwargs.pop('label', None)
        for seq in seqs[1:]:
            plot = TemporalPointSequencePlotter.plot_xy(seq, *args,  **kwargs)
        return plot



from pymeos_cffi import *

from .aggregator import BaseAggregator
from ..boxes import TBox
from ..main import TInt, TFloat, TNumber


class TemporalAverageAggregator(BaseAggregator[TNumber, TNumber]):
    _add_function = tnumber_tavg_transfn
    _final_function = tnumber_tavg_finalfn


class TemporalNumberExtentAggregator(BaseAggregator[TNumber, TBox]):
    _add_function = tnumber_extent_transfn

    @classmethod
    def _finish(cls, state) -> TBox:
        return TBox(_inner=state)


class TemporalIntMaxAggregator(BaseAggregator[TNumber, TBox]):
    _add_function = tint_tmax_transfn


class TemporalIntMinAggregator(BaseAggregator[TInt, TInt]):
    _add_function = tint_tmin_transfn


class TemporalIntSumAggregator(BaseAggregator[TInt, TInt]):
    _add_function = tint_tsum_transfn


class TemporalFloatMaxAggregator(BaseAggregator[TFloat, TFloat]):
    _add_function = tfloat_tmax_transfn


class TemporalFloatMinAggregator(BaseAggregator[TFloat, TFloat]):
    _add_function = tfloat_tmin_transfn


class TemporalFloatSumAggregator(BaseAggregator[TFloat, TFloat]):
    _add_function = tfloat_tsum_transfn



import asyncpg

from pymeos import TBool, TInt, TFloat, TText, TGeomPoint, TGeogPoint, TBox, STBox, TimestampSet, Period, PeriodSet


class MobilityDB:
    @classmethod
    async def connect(cls, *args, **kwargs):
        conn = await asyncpg.connect(*args, **kwargs)
        await cls.register(conn)
        return conn

    @classmethod
    async def register(cls, conn):
        # Add MobilityDB types to PostgreSQL adapter and specify the reader function for each type.
        classes = [TimestampSet, Period, PeriodSet, TBox, TBool, TInt, TFloat, TText, STBox, TGeomPoint, TGeogPoint]
        for cl in classes:
            await conn.set_type_codec(cl.__name__.lower(), encoder=str, decoder=cl.read_from_cursor)
